with Units.Constants ;
with Units.Base ;

with Measures ;
with Measures_Derived ;
with Measures_Irregular;

with Numeric_IO ;
with Numeric_IO.Measure_IO ;
with Name_IO ;
with Name_IO.Measure_Name_IO ;

with Ada.Text_IO;            use Ada.Text_IO ;
with Ada.Integer_Text_IO ;

with Input_Collection;

--  Demonstrate the ability of Name_IO to obtain Kazakov's Measure type
--              Author:         John P Woodruff (jpwoodruff@gmail.com)

procedure Test_Measure_IO is
   type Real is digits 6 ;
   package Real_IO is new Ada.Text_IO.Float_IO (Real) ;
   use Real_IO ;

   package Scalar_IO is new Numeric_IO (Real, Integer ) ;
   use Scalar_IO ;

   package Local_Measures is new Measures(Real);
   use Local_Measures ;

   package Measure_IO is new Scalar_IO.Measure_IO (Local_Measures) ;

   package Local_Measures_Derived is new
     Measures_Derived (Local_Measures) ;

   package Local_Measures_Irregular is new
     Measures_Irregular ( Local_Measures_Derived) ;

   package Name_List is
      type Input_Names_Type is (Name, Force, DX, DY,
                                Distance, Energy, Pressure, Area,
                                A, B, C) ; -- last 3 name unconstrained vars
   end Name_List;

   -- nothing fancy here
   F : Local_Measures.Measure (Units.Constants.Force);
   E : Local_Measures.Measure(Units.Constants.Energy) ;
   P : Local_Measures.Measure(Units.Constants.Pressure) ;
   -- input might have errors
   Distance     : Local_Measures.Measure(Units.Base.Length) ;
   Area         : Local_Measures.Measure(Units.Constants.Area) ;
   -- interact with prompting and default
   DX, DY       :  Local_Measures.Measure(Units.Base.Length)  ;
   --- unconstrained objects
   A, B, C, D   : Local_Measures.Measure ;

   package Nio is new Name_IO (Names_Type    => Name_List.Input_Names_Type,
                               Floating_Type => Real,
                               Integer_Type  => Integer,
                               Scalar_IO     => Scalar_IO);

   package Name_Measure_Input is new
     Nio.Measure_Name_IO (Measure_IO) ;

begin
   --  precondition: this program (might have) been invoked
   --  with a file name on its command line.

   Input_Collection.Read_Collection_Files ;

   --  The file contains a collection of input data for
   --  one instance the object class called "Measures".
   --
   --  this test is provided with an accompanying test data file
   --
   Nio.Acquire_Input ("Measures");

   --  instance may have a name, useful if many in case of interaction
   declare
      Name  : String (1..50);
      Last  : Positive ;
   begin
      Nio.Get (Name_List.Name, Name, Last);
      Nio.Set_Object_Instance_Name (Name (1 .. Last));
   exception
      when Nio.No_Input =>
         Nio.Set_Object_Instance_Name ("Anonymous");
   end;

   Name_Measure_Input.Get (Name_List.Force, F);
   Name_Measure_Input.Get (Name_List.Energy, E) ;
   Name_Measure_Input.Get (Name_List.Pressure, P) ;

   Name_Measure_Input.Get (Name_List.Distance, Distance) ;
   Name_Measure_Input.Get (Name_List.Area, Area, Prompt => "Area?") ;

   Name_Measure_Input.Get
     (Name_List.DX, DX,
      Prompt => "Value for DX (which measures length)? ") ;

   begin
      Name_Measure_Input.Get
        (Name_List.DY, DY) ;
   exception when Nio.No_Input =>
      Put_Line
        ("No value of the input ""DY"" is given, and there is no prompt.") ;
      Put_Line
        ("So the exception handler must provide for correct action.");
      Put_Line ("In the example, ""DY"" is given the value 10 Ft") ;
      DY := 10.0 * Local_Measures_Irregular.Ft;
   end;

   Name_Measure_Input.Get
     (Name_List.A, A,
      Prompt => "Value for Unconstrained measure ""A""");
   Name_Measure_Input.Get
     (Name_List.B, B,
     Prompt => "Value for Unconstrained measure ""B""" );
   Name_Measure_Input.Get
     (Name_List.C, C,
      Prompt => "Value for Unconstrained measure ""C""");


   -- now clear the instance data from NIO state,
   -- in preparation for subsequent instances.
   Nio.Forget_Contents;

   New_Line;
   Put_Line ("Now the test program echoes the data that it has acquired:");
   New_Line;

   Put ("F =>  ") ;  Measure_IO.Put (F) ;
   New_Line;
   Put ("DX =>  ") ;  Measure_IO.Put (DX) ;
   New_Line;
   Put ("DY =>  ") ;  Measure_IO.Put (DY) ;
   New_Line;
   Put ("E =>  ") ;  Measure_IO.Put (E) ;
   New_Line;
   Put ("P =>  ") ;  Measure_IO.Put (P) ;
   New_Line;
   Put ("A =>  ") ;  Measure_IO.Put (A) ;
   New_Line;
   Put ("B =>  ") ;  Measure_IO.Put (B) ;
   New_Line;
   Put ("C =>  ") ;  Measure_IO.Put (C) ;
   New_Line;
   Put ("Distance =>  ") ;  Measure_IO.Put (Distance) ;
   New_Line;
   Put ("Area =>  ") ;  Measure_IO.Put (Area) ;
   New_Line;

end Test_Measure_IO;


-------------------------------------------------------------------------------
--  This is free software; you can redistribute it and/or modify it
--  under terms of the GNU General Public License as published by the
--  Free Software Foundation; either version 2, or (at your option)
--  any later version.  This software is distributed in the hope that
--  it will be useful, but WITH OUT ANY WARRANTY; without even the
--  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
--  PURPOSE. See the GNU General Public License for more details. Free
--  Software Foundation, 59 Temple Place - Suite 330, Boston, MA
--  02111-1307, USA.
--
--  As a special exception, if other files instantiate generics from
--  this unit, or you link this unit with other files to produce an
--  executable, this unit does not by itself cause the resulting
--  executable to be covered by the GNU General Public License. This
--  exception does not however invalidate any other reasons why the
--  executable file might be covered by the GNU Public License.

--  This software was prepared as an account of work sponsored by an
--  agency of the United States Government.  Neither the United States
--  Government nor the University of California nor any of their
--  employees makes any warranty, express or implied, or assumes any
--  liability or responsibility for the accuracy, completeness, or
--  usefulness of any information disclosed.
-------------------------------------------------------------------------------

